(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Interval
 imports
   Main
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
declare [[smt_timeout=120]]
*)

section \<open>Intervals\<close>

text \<open>An interval is a sequence of elements of a particular type. Intervals are similar to list 
   in Isabelle/HOL, the difference is that intervals have instead of nil a single
   element at the end. So an interval of length zero is a single element whereas for Isabelle's list 
   we have that an empty list is nil (no element present). 

   The usual operations on intervals are defined: length (\<open>intlen\<close>), \<open>prefix\<close>, 
   \<open>suffix\<close>, \<open>sub\<close>,  \<open>nth\<close>, \<open>intfirst\<close>, \<open>intlast\<close>, \<open>intapp\<close> 
   and \<open>intrev\<close>. 

   In order to define the semantics of the ITL chopstar we introduce \<open>index_sequence\<close> which
   is a sequence of chop (fuse) points. This sequence is again of type interval but the elements are 
   natural numbers. Two functions \<open>shift\<close> and \<open>shiftm\<close> are introduced that are used to add (shift) and
   subtract a natural number of each element in the sequence of chop (fuse) points.   
    \<close>


subsection \<open>Definitions\<close>

datatype 'a interval =
    St  'a ("\<lceil>_\<rceil>")
  | Cons 'a "'a interval"  (infixr "\<odot>" 65)
for
  map: map
  rel: interval_all2
  pred: interval_all
 
type_synonym  index = "nat interval"  

syntax
  \<comment> \<open>interval Enumeration\<close>
  "_interval" :: "args => 'a interval"    ("\<langle>(_)\<rangle>")

translations
  "\<langle>x, xs\<rangle>" == "x\<odot>\<langle>xs\<rangle>"
  "\<langle>x\<rangle>" == "\<lceil>x\<rceil>"
  
primrec (nonexhaustive) intlen :: "'a interval \<Rightarrow> nat" where
    " intlen (St x) = 0"
  | " intlen (x\<odot>xs) = 1+ (intlen xs)"
 
primrec (nonexhaustive) nth :: "'a interval => nat => 'a" where
    "nth (St x) n      = x"
  | "nth (Cons x xs) n = (case n of 0 \<Rightarrow> x | Suc k \<Rightarrow> nth xs k)"
   
primrec prefix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
    "prefix n (St x) = (St x)"   
  | "prefix n (Cons x xs) = (case n of 0 \<Rightarrow> (St x) | Suc m \<Rightarrow> (Cons x  (prefix m xs)))"

primrec suffix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
   "suffix n (St x) = (St x)" 
 | "suffix n (Cons x  xs) = (case n of 0 \<Rightarrow> (Cons x  xs)  | Suc m \<Rightarrow> suffix m xs)"

definition sub:: "nat \<Rightarrow> nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
where
 "sub n k xs = (if k<n then prefix 0 (suffix n xs)
                       else prefix (k-n) (suffix n xs)
               )" 

primrec intfirst :: "'a interval \<Rightarrow> 'a" where
    "intfirst (St x)     = x"
  | "intfirst (Cons x _) = x"

primrec intlast :: "'a interval \<Rightarrow> 'a" where
    "intlast (St x)      = x"
  | "intlast (Cons _ xs) = intlast xs"  

primrec intapp :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval" (infixr "\<ominus>" 65) where
intapp_St: "(St x) \<ominus> ys = x \<odot> ys" |
intapp_Cons: "(x\<odot>xs) \<ominus> ys = x \<odot> (xs \<ominus> ys)"

primrec intrev :: "'a interval \<Rightarrow> 'a interval" where
  " intrev (St x) = (St x)"
| " intrev (Cons x xs) = (intrev xs) \<ominus> (St x)"

definition index_sequence :: " nat \<Rightarrow> index \<Rightarrow> bool" where
  "index_sequence x idx \<equiv> (nth idx 0 = x) \<and> (\<forall> n. n<intlen idx \<longrightarrow> nth idx n < nth idx (Suc n))"  

definition shift :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shift k = (\<lambda> x. x+k)"

definition shiftm ::  "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shiftm k = (\<lambda> x. (if k>x then 0 else  (x-k)))" 



subsection \<open>Lemmas\<close>

text \<open>Basic lemmas are introduced for each of the above operations on intervals. \<close>
  

subsubsection \<open>Interval Length\<close>

lemma interval_intlen_gr_zero [simp]:
    "intlen xs \<ge> 0"
by auto

lemma interval_intlen_st :
    " intlen (St x) = 0" 
by simp

lemma interval_intlen_cons [simp]: 
    " (intlen (x\<odot>xs)) = (intlen xs) +1"
by simp
  
lemma interval_intlen_cons_1 :
    " intlen l > 0 \<longleftrightarrow> (\<exists> x ls. l = x \<odot>ls) "
by (induct l) simp_all

lemma interval_intlen_map:
    " intlen (map f xs) = intlen xs"
by (induct xs) simp_all



subsubsection \<open>nth\<close>

lemma interval_nth_zero [simp]:
    " nth (x\<odot>xs) 0  = x"
by simp

lemma interval_nth_Suc [simp]:
    "nth (x\<odot>xs) (Suc n) = nth xs n"
by auto

lemma interval_nth_last:
    " nth (x\<odot>xs) (intlen (x\<odot>xs)) = nth xs (intlen xs) "
by simp

lemma interval_nth_cons:
 assumes "0<i \<and> i<1+intlen(xs)"
 shows   "nth(x\<odot>xs) i =nth xs (i-1) \<and>
          nth(x\<odot>xs) (i+1) = nth xs ((i-1)+1)"
by (metis One_nat_def Suc_leI add.commute assms interval_nth_Suc le_add_diff_inverse2 plus_1_eq_Suc)

lemma interval_nth_zero_intfirst:
 " nth xs 0 = intfirst xs "
by (induct xs) simp_all

lemma interval_nth_intlen_intlast:
 " nth xs (intlen xs) = intlast xs"
by (induct xs) simp_all

lemma interval_st_intlen :
    " (xs = (St x)) \<longleftrightarrow> intlen xs = 0 \<and> nth xs 0 = x" 
by (induct xs) simp_all

lemma interval_eq_nth_eq :
    " (xs =ys) = (intlen xs =intlen ys \<and> (\<forall> i\<le> intlen xs. nth xs i = nth ys i))"
apply (induct xs arbitrary: ys)
apply (metis interval_st_intlen le_numeral_extra(3))
apply (case_tac ys, simp)
by fastforce

lemma interval_nth_map :
    "nth (map f xs) i = f (nth xs i)"
apply (induct xs arbitrary: i, simp)
apply (case_tac i, simp, simp)
done

subsubsection \<open>index sequence\<close>

lemma interval_idx_less:
    assumes iseq: " index_sequence x idx"
    shows " (n<intlen idx \<and> n+k<intlen idx) \<longrightarrow> nth idx n < nth idx (Suc(n+k)) "
apply (induct k)
using index_sequence_def iseq apply auto[1]
using index_sequence_def iseq by auto

lemma interval_idx_less_last :
 assumes " index_sequence x idx"
 shows   " (i<intlen idx \<and> i+(intlen idx - (i+1))< intlen idx)
            \<longrightarrow> nth idx i < nth idx (Suc(i+(intlen idx -(i+1))))"
using assms interval_idx_less by blast

lemma interval_idx_less_last_1:
 assumes  "index_sequence x idx"
 shows    "i<intlen idx \<longrightarrow> nth idx i < nth idx (intlen idx) "
using assms interval_idx_less_last by auto

lemma interval_idx_greater_first:
 assumes  " index_sequence x idx"
 shows    " (i>0 \<and> i\<le> intlen idx) \<longrightarrow> x < nth idx i"
apply (induct i, simp)
using assms 
by (metis One_nat_def Suc_le_lessD add_Suc index_sequence_def interval_idx_less 
          less_le_trans plus_1_eq_Suc)

lemma interval_idx_cons:
    " index_sequence 0 (x\<odot>ls) = 
       (x=0 \<and> x<nth ls 0 \<and> index_sequence (nth ls 0) ls)"
apply (simp add: index_sequence_def)
using less_Suc_eq_0_disj by auto

lemma interval_idx_shift_mono:
    "mono (shift k)" 
by (simp add: Interval.shift_def mono_def)     

lemma interval_idx_expand:
 "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 0\<le>i \<and> i< (intlen l)
  \<Longrightarrow> 0\<le> (nth l i) \<and> (nth l i) \<le>  (nth l (i+1)) \<and> (nth l (i+1)) \<le> (intlen xs)"
apply (simp add: index_sequence_def)
apply (induct l, simp)
by (metis Suc_lessI eq_imp_le index_sequence_def interval_idx_less_last_1 less_imp_le_nat)

lemma interval_idx_shift_idx [simp]:
    " ( index_sequence (x+k) (map (shift k)  idx)) = (index_sequence x idx) " 
by (simp add: Interval.shift_def index_sequence_def interval_intlen_map interval_nth_map)    

lemma interval_idx_shiftm :
    "(index_sequence k (lsk) \<and> ls = map (shiftm k) lsk) \<Longrightarrow> 
     index_sequence 0 (ls) \<and> (intlen ls) = (intlen lsk)"
by (simp add: interval_eq_nth_eq index_sequence_def shift_def shiftm_def interval_nth_map ) 
   (smt Suc_leI diff_less_mono index_sequence_def interval_idx_greater_first interval_intlen_map 
        le_less_trans less_Suc_eq_0_disj not_less order.asym)

lemma interval_lsk_ls :
    "(index_sequence k (lsk) \<and> lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
      (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> index_sequence 0 (ls) ) "
apply (simp add: interval_eq_nth_eq index_sequence_def shift_def shiftm_def interval_nth_map)
apply rule
apply (metis (no_types, lifting) add_diff_cancel_right' interval_intlen_map not_add_less2)
by (metis (no_types, lifting) Suc_eq_plus1 add.commute add_cancel_right_left add_diff_inverse_nat 
       ex_least_nat_less interval_intlen_map le_SucE le_zero_eq not_less_zero order_refl)

lemma interval_idx_link_shiftm:
    "(index_sequence k (lsk) \<and> ls = map (shiftm k) lsk ) =
     (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> 
      index_sequence 0 (ls) \<and> (intlen ls) =(intlen lsk)) "
using interval_idx_shiftm by blast
    
lemma interval_idx_link:
    "(lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
     (lsk = map (shift k) ls \<and> index_sequence k (lsk) \<and> index_sequence 0 (ls)\<and> 
      (intlen ls) =(intlen lsk))" 
by (metis Interval.shift_def add_diff_cancel_left' diff_diff_cancel diff_is_0_eq' 
    interval_idx_shift_idx interval_idx_shift_mono interval_intlen_map le_numeral_extra(3) mono_def)

lemma interval_idx_bound_0 :
 assumes "index_sequence 0 ls \<and> Interval.nth ls (intlen ls) = intlen (suffix k xs)"
 shows   "((i\<le>intlen ls) \<longrightarrow> ((nth ls (i)) \<le> (intlen (suffix k xs))))"
using assms 
by (metis add.commute add_eq_if eq_iff interval_idx_less le_add_diff_inverse2 
                le_neq_implies_less lessI less_imp_le_nat)
  
lemma interval_idx_bound_1:
  "(index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs))) \<longleftrightarrow>
     (index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs)) \<and>
     (\<forall>i. (i\<le>intlen ls) \<longrightarrow> ((nth ls (i)) \<le> (intlen (suffix k xs)))) )" 
using interval_idx_bound_0 by blast

subsubsection \<open>prefix, suffix and sub\<close>

lemma interval_prefix_state [simp]: 
    " prefix m (St x) = (St x) " 
by simp

lemma interval_prefix_suc [simp]:
    " prefix (Suc m) (x\<odot>xs) = x \<odot> (prefix m xs)" 
by auto

lemma interval_prefix_zero [simp]:
    " prefix 0 (x\<odot>xs) = St x"
by auto

lemma interval_prefix_zero_intfirst [simp]:
  " prefix 0 xs = St (intfirst xs)"
by (induct xs) simp_all

lemma interval_intfirst_prefix [simp]:
 " i\<le>intlen xs \<Longrightarrow>   intfirst (prefix i xs) = intfirst xs"
by (induct xs arbitrary: i, auto) (case_tac i, auto)
  
lemma interval_prefix_intlen [simp]:
    " (prefix (intlen xs) xs) = xs"
by (induct xs) simp_all

lemma interval_prefix_intlen_gr_1 [simp]: 
    "(prefix ((intlen xs)+i) xs) = xs"
by (induct xs) simp_all

lemma interval_intlen_prefix_cons [simp]:
    " intlen( prefix (Suc i) (x\<odot>xs)) = 1 + intlen(prefix i xs)"
using interval_intlen_cons by auto

lemma interval_prefix_length : 
    " intlen (prefix i xs) = (if i\<le> intlen xs then i else intlen xs)"
by (induct xs arbitrary: i, simp) (case_tac i, auto) 

lemma interval_prefix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " (intlen ( prefix i xs)) = i" 
using assms by (simp add: interval_prefix_length)

lemma interval_prefix_length_bad [simp] : 
  assumes   " i > intlen xs"
  shows     "  intlen (prefix i xs) = intlen xs"
using assms by (simp add: interval_prefix_length)
   
lemma interval_pref_intlen_bound :
 assumes   " i \<le> (intlen xs)"
 shows     "  intlen (prefix i xs) \<le> intlen xs " 
using assms by (induct xs, simp) (metis interval_prefix_length)

lemma interval_suffix_length:
    " intlen (suffix i xs ) = (if i\<le> intlen xs then (intlen xs)-i else 0) "
by (induct xs arbitrary: i, simp) (case_tac i, auto)
    
lemma interval_suffix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " intlen (suffix i xs ) = (intlen xs)-i"
using assms by (simp add: interval_suffix_length)
  
lemma interval_suffix_length_bad [simp]:
  assumes  " i> intlen xs"
  shows    "  intlen (suffix i xs ) = 0"
using assms by (simp add: interval_suffix_length)
 
lemma interval_nth_prefix [simp]:
 " i\<le>intlen xs \<and> k \<le> i \<Longrightarrow> nth (prefix i xs) k = nth xs k"
apply (induct xs arbitrary: i k, auto)
apply (case_tac i, auto) 
apply (case_tac k, auto)
done

lemma interval_nth_suffix [simp]:
 " i\<le>intlen xs \<and> k \<le> intlen xs -i \<Longrightarrow> nth (suffix i xs) k = nth xs (i+k)"
by (induct xs arbitrary: i k, auto) (case_tac i, auto)

lemma interval_suffix_prefix_help_1:
 assumes " ia+i \<le> intlen xs \<and> k \<le> ia"
 shows " nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k"
proof -
 have 1: " nth (prefix ia (suffix i xs)) k = nth (suffix i xs) k" 
 using interval_nth_prefix assms by (metis interval_prefix_intlen_gr_1 le_cases le_iff_add) 
 have 2: " nth (suffix i xs) k = nth xs (i+k) " 
 using interval_nth_suffix assms by (simp add: add_le_imp_le_diff)
 have 3: " nth xs (i+k) = nth (prefix (ia+i) xs) (i+k)" 
 using interval_nth_prefix assms by simp
 have 4: " nth (prefix (ia+i) xs) (i+k) = nth (suffix i (prefix (ia +i) xs)) k" 
 using interval_nth_suffix assms by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma interval_suffix_prefix_help_2:
 assumes "ia+i \<le> intlen xs"
 shows   " (\<forall> k \<le> ia . nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k)"
using interval_suffix_prefix_help_1 using assms by fastforce

lemma interval_suffix_prefix_help_3:
 assumes "ia+i \<le> intlen xs "
 shows   "intlen (prefix ia (suffix i xs)) = intlen (suffix i (prefix (ia +i) xs)) "
using assms interval_prefix_length_good interval_suffix_length_good by auto

lemma interval_suffix_prefix_swap:
 assumes "ia+i \<le> intlen xs"
 shows   " prefix ia (suffix i xs) = suffix i (prefix (ia +i) xs)"
by (simp add: interval_eq_nth_eq interval_suffix_prefix_help_2 interval_suffix_prefix_help_3 assms)
 
lemma interval_prefix_prefix_zero [simp]:
    " prefix 0 ( prefix 0 xs ) = prefix 0 xs" 
by (induct xs) simp_all
  
lemma interval_pref_pref [simp]:
    " (prefix i (prefix i xs)) = prefix i xs"
by (metis interval_prefix_intlen interval_prefix_intlen_gr_1 interval_prefix_length 
          less_imp_add_positive not_less)

lemma interval_pref_pref_3 [simp]:
    " (prefix i (prefix (i+k) xs)) = prefix i xs"
apply (induct xs arbitrary: i k, simp)
apply (case_tac i, auto) 
by (simp add: Nitpick.case_nat_unfold)
 
lemma interval_pref_help:
 assumes "i\<le>intlen (prefix (intlen xs - Suc 0) xs)"
 shows   " (prefix i (prefix (intlen xs - Suc 0) xs)) = (prefix i xs)  "
using assms 
by (metis diff_le_self interval_pref_pref_3 interval_prefix_length 
           ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma interval_pref_pref_help:
 assumes "intlen xs >0 \<and> ia<intlen (xs)"
 shows   "(prefix ia (prefix (intlen xs - Suc 0) xs)) = (prefix ia  xs) "
using assms 
by (metis Suc_leI Suc_le_mono Suc_pred diff_le_self interval_pref_help interval_prefix_length_good)

lemma interval_pref_pref_help_1:
 assumes "i>0 \<and> i\<le> intlen xs"
 shows   "(prefix (intlen (prefix i xs) - Suc 0) (prefix i xs)) = 
          (prefix (intlen (prefix i xs) - Suc 0) xs) "
using assms interval_pref_pref_3 by (metis diff_le_self interval_prefix_length_good le_iff_add)

lemma interval_suffix_suc [simp]:
    " suffix (Suc m) (x \<odot> xs) = suffix m xs"
by auto
  
lemma interval_suffix_zero [simp]:
    " suffix 0 xs = xs "
by (induct xs) simp_all
    
lemma interval_suffix_intlen [simp]:
    " suffix (intlen xs) xs = (St (nth xs (intlen xs))) "
by (induct xs) simp_all

lemma interval_suffix_intlast [simp]:
   " suffix (intlen xs) xs = St (intlast xs)"
by (induct xs) simp_all

lemma interval_suffix_suffix [simp]:
    " suffix i (suffix j xs) = suffix (i+j) xs "
apply (induct xs arbitrary: i j, simp)
apply (case_tac i, auto)
by (simp add: Nitpick.case_nat_unfold)

lemma interval_prefix_suffix_intlen:
    "intlen (prefix ia (suffix i xs)) =
     (if i \<le> intlen xs then
      (if ia\<le> intlen xs -i  then ia else (intlen xs) -i )
      else 0)"
by (metis interval_prefix_length interval_suffix_length le_zero_eq)

lemma interval_prefix_suffix_intlen_good [simp]:
  assumes  "ia\<le> intlen xs -i \<and> i \<le> intlen xs"
  shows    " intlen (prefix ia (suffix i xs)) = ia "
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_prefix_suffix_intlen_bad_0 [simp]:
  assumes  "i> intlen xs"
  shows    "intlen (prefix ia (suffix i xs)) = 0"
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_prefix_suffix_intlen_bad_1 [simp] :
  assumes  "i \<le> intlen xs \<and> ia > intlen xs -i"
  shows    "intlen (prefix ia (suffix i xs)) = (intlen xs) -i  "
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_suffix_suffix_3:
 assumes "i>0 \<and> ia<i \<and> i\<le> intlen xs"
 shows   "(suffix (i-ia) (suffix ((intlen xs)-i) xs)) = (suffix (((intlen xs)-ia)) xs)"
using assms by simp

lemma interval_sub_zero_prefix :
    " sub 0 k xs = prefix k xs"
by (simp add: Interval.sub_def) 

lemma interval_sub_suffix : 
  assumes  "(i < j \<and> j\<le> (intlen xs)-k)"
  shows    "(sub (i+k) (j+k) xs) = (sub i j (suffix k xs))"
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix_0:
  assumes "(0 \<le> i \<and> ia+i \<le> intlen xs)"
  shows   "  (sub i (i+ia) xs) = (prefix (ia) (suffix i xs))" 
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   "  (sub i j xs) = (prefix (j-i) (suffix i xs))"
using assms by (simp add: Interval.sub_def) 


subsubsection \<open>Reverse\<close>

lemma interval_intlen_intapp [simp]:
 " intlen (xs \<ominus> ys) = (intlen xs) + (intlen ys) +1"
by (induct xs arbitrary: ys) simp_all

lemma interval_intrev_intlen [simp]:
 " intlen (intrev xs) = intlen xs"
by (induct xs, simp, simp)

lemma interval_suffix_intapp [simp]:
 "suffix (Suc (intlen xs)) (xs \<ominus> ys) = ys "
by (induct xs) simp_all

lemma interval_suffix_intapp2 [simp]: 
 "suffix (intlen xs - k) ( xs) \<ominus> ys = suffix (intlen xs - k) ( xs \<ominus> ys)"
by (induct xs, simp)
   (metis Suc_diff_le diff_is_0_eq' intapp_Cons interval_suffix_suc interval_suffix_zero 
          intlen.simps(2) not_less_eq_eq plus_1_eq_Suc)

lemma interval_intapp_assoc [simp]:
 " (xs \<ominus> ys) \<ominus> zs = xs \<ominus> ( ys \<ominus> zs ) "
by (induct xs) simp_all

lemma interval_intapp_nth:
 " nth (xs \<ominus> ys) k = (if k \<le> intlen xs 
                      then (nth xs k) 
                      else (nth ys (k - (intlen xs) -1)) )"
apply (induct xs arbitrary: k)
apply (case_tac k, simp, simp) 
apply (case_tac k, simp, simp)
done

lemma interval_rev_intapp [simp]:
 " intrev (xs \<ominus> ys) = (intrev ys) \<ominus> (intrev xs)"
by (induct xs) simp_all

lemma interval_rev_rev_ident [simp]: 
 " intrev (intrev xs) = xs"
by (induct xs) auto

lemma interval_rev_swap :
 " ((intrev xs) = ys) = (xs = intrev ys)"
by auto

lemma interval_intlast_intrev:
 "intlast (intrev xs) = intfirst xs"
by (induct xs, auto)
   (metis Suc_eq_plus1 add.right_neutral interval.inject(1) interval_intlen_intapp 
          interval_intlen_st interval_suffix_intapp interval_suffix_intlast)
 
lemma interval_intfirst_intrev:
 "intfirst (intrev xs) = intlast xs"
by (induct xs, auto)
   (metis intapp_St interval_intlast_intrev interval_rev_intapp intlast.simps(2) intrev.simps(1))

lemma interval_intrev_nth:
 "k\<le> intlen (intrev xs) \<Longrightarrow> (nth  (intrev xs) k) = (nth  xs ((intlen xs) -k))"
apply (induct xs, simp)
apply simp
apply (case_tac k)
apply (simp add: interval_intapp_nth)
by (smt Interval.nth.simps(1) Suc_diff_Suc diff_Suc_Suc diff_is_0_eq' interval_intapp_nth 
        interval_intrev_intlen le_SucE less_Suc_eq_le old.nat.simps(4) old.nat.simps(5))

lemma interval_intrev_prefix:
 "k\<le> intlen xs \<Longrightarrow> intrev( prefix k xs) = suffix ((intlen xs) - k) (intrev xs)  "
apply (induct xs arbitrary: k, simp)
apply simp
apply (case_tac k)
apply (metis diff_zero interval_intrev_intlen interval_suffix_intapp intrev.simps(1) old.nat.simps(4))
by (metis Suc_le_mono diff_Suc_Suc interval_intrev_intlen interval_suffix_intapp2 
          intrev.simps(2) old.nat.simps(5))

lemma interval_intrev_suffix:
 "k\<le> intlen xs \<Longrightarrow> intrev( suffix k xs) = prefix ((intlen xs) - k) (intrev xs)  "
by (induct xs arbitrary: k, simp, simp add: interval_intrev_prefix interval_rev_swap)

lemma interval_intrev_sub1:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   " intrev (sub i j xs) = intrev (prefix (j-i) (suffix i xs)) "
using assms interval_sub_prefix_suffix by (simp add: interval_sub_prefix_suffix)

lemma interval_intrev_sub2:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   " intrev (prefix (j-i) (suffix i xs)) = suffix ((intlen xs) - j) (intrev (suffix i xs))"
using assms interval_intrev_prefix[of "j-i" "suffix i xs"]  by auto

lemma interval_intrev_sub3:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   " suffix ((intlen xs) - j) (intrev (suffix i xs)) = 
           suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs))"
using assms interval_intrev_suffix[of "i" "xs"] by auto

lemma interval_intrev_sub4:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   "suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs)) =
          sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs)"
using assms by (simp add: diff_le_mono2 interval_sub_prefix_suffix interval_suffix_prefix_swap)

lemma interval_intrev_sub:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   " intrev (sub i j xs) = sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs)"
using assms 
by (simp add: interval_intrev_sub1 interval_intrev_sub2 interval_intrev_sub3 interval_intrev_sub4)

lemma interval_intrev_idx_2:
 assumes "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and>
          0\<le>i \<and> i< (intlen l)"
 shows  " (intrev (sub (nth l i) (nth l (i+1)) xs)) =
   ( (sub ((intlen xs)- (nth l (i+1))) ((intlen xs) - (nth l i)) (intrev xs)))"
using assms interval_idx_expand interval_intrev_sub[of "(nth l i) " "(nth l (i+1)) " "xs"] 
by blast

lemma interval_intrev_idx_3:
 assumes " index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
           ls = map (\<lambda> x. (intlen xs) -x) (intrev l)"
 shows   "(nth ls 0) = 0 \<and> (nth ls (intlen ls)) = (intlen xs) \<and> intlen ls =intlen l"
using assms 
by (metis diff_self_eq_0 diff_zero index_sequence_def interval_intfirst_intrev 
          interval_intlast_intrev interval_intlen_map interval_intrev_intlen 
          interval_nth_intlen_intlast interval_nth_map interval_nth_zero_intfirst)

lemma interval_intrev_idx_4:
 " index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
             ls = map (\<lambda> x. (intlen xs) -x) (intrev l)
   \<Longrightarrow> i\<le> intlen ls \<longrightarrow> (nth ls i) = (intlen xs) - (nth l ((intlen l)-i)) "
apply (induct ls)
apply (metis diff_zero interval_intlen_st interval_intrev_idx_3 le_0_eq)
by (simp add: interval_intlen_map interval_intrev_nth interval_nth_map)

lemma interval_intrev_idx_5:
 assumes " (index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs))"
 shows   " (i< intlen l \<longrightarrow> 
           (intlen xs) - (nth l ((intlen l)-i)) < (intlen xs) - (nth l ((intlen l)-(i+1)))) "
using assms 
by (smt Suc_diff_Suc Suc_eq_plus1 add_gr_0 add_less_cancel_left diff_less 
        index_sequence_def le_add_diff_inverse2 le_numeral_extra(3) less_diff_conv 
        less_imp_le_nat not_gr_zero interval_idx_expand)

lemma interval_intrev_idx_6:
 assumes " (index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
           ls = map (\<lambda> x. (intlen xs) -x) (intrev l))"
 shows   " (i< intlen ls \<longrightarrow> 
           ((nth ls i) = (intlen xs) - (nth l ((intlen l)-i)) \<and>
           (nth ls (i+1)) = (intlen xs) - (nth l ((intlen l)-(i+1))) \<and>
           (nth ls i) < (nth ls (i+1))))"
proof -
 have 1: "(i< intlen ls \<longrightarrow> (nth ls i) = (intlen xs) - (nth l ((intlen l)-i))) " 
          using assms interval_intrev_idx_4 less_imp_le_nat by blast
 have 2: "(i< intlen ls \<longrightarrow> (nth ls (i+1)) = (intlen xs) - (nth l ((intlen l)-(i+1)))) " 
          using assms by (simp add: interval_intrev_idx_4)
 have 3: "( i< intlen ls \<longrightarrow> 
           ((nth ls i) = (intlen xs) - (nth l ((intlen l)-i)) \<and>
           (nth ls (i+1)) = (intlen xs) - (nth l ((intlen l)-(i+1)))))
         " using "1" "2" by auto
 have 4: "(i< intlen ls \<longrightarrow> 
             ((nth ls i) = (intlen xs) - (nth l ((intlen l)-i)) \<and>
             (nth ls (i+1)) = (intlen xs) - (nth l ((intlen l)-(i+1))) \<and>
             (nth ls i) < (nth ls (i+1))))
          " using assms "3" index_sequence_def interval_intrev_idx_5
          by (metis interval_intlen_map interval_intrev_intlen)
 from 4 show ?thesis by blast
qed

lemma interval_intrev_idx_7:
 assumes " (index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
           ls = map (\<lambda> x. (intlen xs) -x) (intrev l))"
   shows " index_sequence 0 ls "
using assms interval_intrev_idx_6 interval_intrev_idx_3 
by (metis Suc_eq_plus1 index_sequence_def)

lemma interval_intrev_idx_8:
 assumes "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
          ls = map (\<lambda> x. (intlen xs) -x) (intrev l) \<and> index_sequence 0 ls "
 shows   "i<intlen ls \<longrightarrow> 
          (intlen xs)- (nth l (i+1)) = nth ls ((intlen ls)-(i+1)) \<and>
          (intlen xs) - (nth l i)    = (nth ls ((intlen ls) -i))"
using assms interval_intrev_idx_4 
by (smt Suc_eq_plus1 Suc_leI add_diff_cancel_right' assms diff_diff_cancel diff_diff_left 
        diff_le_self interval_intrev_idx_3)

lemma interval_intrev_idx_9:
 assumes "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 
          ls = map (\<lambda> x. (intlen xs) -x) (intrev l) \<and> index_sequence 0 ls "
 shows   "i<intlen ls \<longrightarrow> 
          sub ((intlen xs)- (nth l (i+1))) ((intlen xs) - (nth l i)) (intrev xs) =
          sub (nth ls ((intlen ls)-(i+1))) ((nth ls ((intlen ls) -i)) ) (intrev xs) 
          "
using interval_intrev_idx_8 using assms by fastforce

lemma interval_intrev_idx_11:
 assumes "(index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs))"
 shows   " i\<le>intlen l \<longrightarrow>
         (nth l i) = (nth (map (\<lambda> x.(intlen xs)-x) (intrev (map (\<lambda> x.(intlen xs)-x) (intrev l)))) i)"
using assms index_sequence_def 
by (smt diff_diff_cancel diff_is_0_eq diff_less diff_zero leD le_cases not_gr_zero 
        interval_intrev_idx_3 interval_intrev_idx_6 interval_intrev_idx_7)

lemma interval_intrev_idx_12:
 assumes "(index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs))" 
 shows   " l = map (\<lambda> x. (intlen xs) -x  ) (intrev (map (\<lambda> x. (intlen xs) -x) (intrev l))) "
using assms interval_intrev_idx_11 
by (simp add: interval_intrev_idx_11 interval_eq_nth_eq interval_intlen_map)


end
