(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Monitor
imports First

begin

section \<open>Monitors\<close>

text \<open>
The RV monitors language is  introduced plus the algebraic properties of the monitor operators. 
\<close>

subsection \<open>Syntax\<close>

datatype 'a monitor =
   mFIRST_d "'a pitl" ("(FIRST _)")
 | mUPTO_d "'a monitor" "'a monitor" ("(_ UPTO _)" [84,84] 83)
 | mTHRU_d "'a monitor" "'a monitor" ("(_ THRU _)" [84,84] 83)
 | mTHEN_d "'a monitor" "'a monitor" ("(_ THEN _)" [84,84] 83)
 | mWITH_d "'a monitor" "'a pitl"    ("(_ WITH _)" [84,84] 83)

subsection \<open>Derived Monitors\<close>

definition mHALT_d ("(HALT _)" [84] 83)
where
 "HALT w \<equiv> FIRST (fin (init w))"

definition mLEN_d ("(LEN _)" [84] 83)
where
 "LEN k \<equiv> FIRST (len k)"

definition mEMPTY_d ("EMPTY")
where
 "EMPTY \<equiv> FIRST empty"

definition mSKIP_d ("SKIP")
where
 "SKIP \<equiv> FIRST skip"

definition mGUARD_d ("(GUARD _)" [84] 83)
where
 "GUARD w \<equiv> EMPTY WITH ( (init w))"  

definition mFAIL_d ("FAIL")
where 
 "FAIL \<equiv> GUARD false\<^sub>i"

primrec mTIMES_d :: "'a monitor \<Rightarrow> nat \<Rightarrow> 'a monitor" ("(_ TIMES _)" [84,84] 83)
where
  mTIMES_0  : "a TIMES 0       = EMPTY"
| mTIMES_Suc: "a TIMES (Suc k) = a THEN (a TIMES k)"

definition mALWAYS_d ("(_ ALWAYS _)" [84,84] 83)
where 
 "a ALWAYS (w) \<equiv> a WITH (bi (fin (init w)))"

definition mSOMETIME_d ("(_ SOMETIME _)" [84,84] 83)
where 
 "a SOMETIME (w) \<equiv> a WITH (di (fin (init w)))"
 
definition mlimit_d ("(Limit _)" [84] 83)
where 
 "Limit f \<equiv> (bs (\<not>\<^sub>i f))"  (*  \<equiv> (\<not>\<^sub>i di (f;skip))  *)

definition mWITHIN_d ("(_ WITHIN _)" [84,84] 83)
where 
 "a WITHIN (f) \<equiv> a WITH (Limit f)"

definition mUNTIL_d ("(_ UNTIL _)" [84,84] 83)
  where
 "w1 UNTIL w2 \<equiv> (HALT w2) WITH (bm w1)"

subsection \<open>Semantics\<close>

fun semantics_monitor :: "'a monitor \<Rightarrow> 'a pitl" ("(\<M> _)" [80] 80)
where
  "(\<M> (FIRST a))  = \<rhd> a"
| "(\<M> (a UPTO b)) = \<rhd>( (\<M> a) \<or>\<^sub>i (\<M> b))"
| "(\<M> (a THRU b)) = \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> b))"
| "(\<M> (a THEN b)) = ((\<M> a);(\<M> b))"
| "(\<M> (a WITH f)) = ( (\<M> a) \<and>\<^sub>i f )" 


definition eq_d :: "'a monitor \<Rightarrow> 'a monitor \<Rightarrow> bool" ("(_ \<simeq> _)" [84,84] 83)
where
 "eq_d a b \<equiv> (\<turnstile> (\<M> a) \<equiv>\<^sub>i (\<M> b))"

lemma MonEqRefl:
 "a \<simeq> a"
by (simp add: eq_d_def)

lemma MonEqSym:
 assumes "a \<simeq> b"
 shows   "b \<simeq> a"
using assms eq_d_def itl_prop(30) by blast

lemma MonEqTrans:
 assumes "a \<simeq> b"
         "b \<simeq> c"
 shows   "a \<simeq> c"
using assms(1) assms(2) using eq_d_def prop03 by blast

lemma MonEq:
 "(a \<simeq> b) = (\<turnstile> (\<M> a) \<equiv>\<^sub>i (\<M> b))"
by (simp add: eq_d_def)

lemma MonEqSubstWith:
 assumes "a \<simeq> b"
 shows   "(a WITH f) \<simeq> (b WITH f)"
using assms by (simp add: eq_d_def)

lemma MonEqSubstThen:
 assumes "a1 \<simeq> b1"
         "a2 \<simeq> b2"
 shows   "(a1 THEN a2) \<simeq> (b1 THEN b2)"
using assms(1) assms(2) by (simp add: eq_d_def)

lemma MonEqSubstUpto:
 assumes "a1 \<simeq> b1"
         "a2 \<simeq> b2"
 shows   "(a1 UPTO a2) \<simeq> (b1 UPTO b2)"
using assms(1) assms(2) 
proof -
 have 1: "a1 \<simeq> b1" using assms(1) by blast
 have 2: "a2 \<simeq> b2" using assms(2) by blast
 have 3: "((a1 UPTO a2) \<simeq> (b1 UPTO b2)) =
          (\<turnstile> (\<M> a1 UPTO a2) \<equiv>\<^sub>i (\<M> b1 UPTO b2))" by (simp add: eq_d_def)
 have 4: "\<turnstile>  (\<M> a1 UPTO a2) \<equiv>\<^sub>i \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2))" by simp
 have 5: "\<turnstile> (\<M> b1 UPTO b2) \<equiv>\<^sub>i \<rhd>( (\<M> b1) \<or>\<^sub>i (\<M> b2))" by simp
 have 6: "\<turnstile> ( (\<M> a1) \<or>\<^sub>i (\<M> a2)) \<equiv>\<^sub>i ( (\<M> b1) \<or>\<^sub>i (\<M> b2))" using "1" "2" eq_d_def by auto 
 have 7: "\<turnstile> \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2)) \<equiv>\<^sub>i \<rhd>( (\<M> b1) \<or>\<^sub>i (\<M> b2))" using "6" FstEqvRule by blast
 have 8: " (\<turnstile> (\<M> a1 UPTO a2) \<equiv>\<^sub>i (\<M> b1 UPTO b2))" using "7" "6" "5" by auto
 from 3 7 show ?thesis by auto
qed 

lemma MonEqSubstThru:
 assumes "a1 \<simeq> b1"
         "a2 \<simeq> b2"
 shows   "(a1 THRU a2) \<simeq> (b1 THRU b2)"
using assms(1) assms(2) 
proof -
 have 1: "a1 \<simeq> b1" 
      using assms(1) by blast
 have 2: "a2 \<simeq> b2" 
      using assms(2) by blast
 have 3: "((a1 THRU a2) \<simeq> (b1 THRU b2)) =
          (\<turnstile> (\<M> a1 THRU a2) \<equiv>\<^sub>i (\<M> b1 THRU b2))" 
      by (simp add: eq_d_def)
 have 4: "\<turnstile>  (\<M> a1 THRU a2) \<equiv>\<^sub>i \<rhd>( di(\<M> a1) \<and>\<^sub>i di(\<M> a2))" 
      by simp
 have 5: "\<turnstile> (\<M> b1 THRU b2) \<equiv>\<^sub>i \<rhd>( di(\<M> b1) \<and>\<^sub>i di(\<M> b2))" 
      by simp
 have 6: "\<turnstile> ( (\<M> a1) \<and>\<^sub>i (\<M> a2)) \<equiv>\<^sub>i ( (\<M> b1) \<and>\<^sub>i (\<M> b2))" 
      using "1" "2" eq_d_def by auto 
 have 7: "\<turnstile> ( di(\<M> a1) \<and>\<^sub>i di(\<M> a2)) \<equiv>\<^sub>i ( di(\<M> b1) \<and>\<^sub>i di(\<M> b2))" 
      using "6" by (meson "1" "2" DiEqvDi eq_d_def itl_prop(31) prop22)
 have 8: "\<turnstile> \<rhd>( di(\<M> a1) \<and>\<^sub>i di(\<M> a2)) \<equiv>\<^sub>i \<rhd>( di(\<M> b1) \<and>\<^sub>i di(\<M> b2))"  
      using "7" FstEqvRule by blast
 have 9: " (\<turnstile> (\<M> a1 THRU a2) \<equiv>\<^sub>i (\<M> b1 THRU b2))" 
      using "8" "5" "4" by auto
 from 3 9 show ?thesis by auto
qed 

definition mAND_d ("(_ AND _)" [84,84] 83)
where 
  " a AND b \<equiv> a WITH (\<M> b) "

definition mITERATE_d ("(_ ITERATE _)" [84,84] 83)
where
  " a ITERATE b \<equiv> a WITH (\<M> b)\<^sup>\<star>"

definition mSTAR_d ("(_ STAR _)" [84,84] 83)
where
 " a STAR f \<equiv> (FIRST(\<diamond> f)) ITERATE (a)"

definition mREPEAT_d ("(_ REPEATUNTIL _)" [84,84] 83)
where 
  " a REPEATUNTIL w \<equiv> ( (HALT w) ITERATE (a WITH (keep(\<not>\<^sub>i (init w)))))"


subsection \<open>Monitor Laws\<close>

lemma MFixFst:
 "\<turnstile> (\<M> a) \<equiv>\<^sub>i \<rhd> (\<M> a)"
proof
 (induct a )
 case (mFIRST_d x)
 then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> FIRST x) \<equiv>\<^sub>i \<rhd> x"  by simp
   have 2: "\<turnstile> \<rhd> x \<equiv>\<^sub>i \<rhd> (\<rhd> x)" using FstFixFst by auto
   have 3: "\<turnstile>  \<rhd> (\<rhd> x) \<equiv>\<^sub>i \<rhd>(\<M> FIRST x)" by simp
   from 1 2 3 show ?thesis by auto
  qed
 next
  case (mUPTO_d a1 a2)
  then show ?case 
  proof -
  have 1: "\<turnstile> (\<M> (a1 UPTO a2)) \<equiv>\<^sub>i \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2))" by simp
  have 2: "\<turnstile>  \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<rhd>((\<M> a1) \<or>\<^sub>i (\<M> a2)))" using FstFixFst by auto
  have 3: "\<turnstile> \<rhd>(\<rhd>((\<M> a1) \<or>\<^sub>i (\<M> a2))) \<equiv>\<^sub>i \<rhd>(\<M> (a1 UPTO a2))" by simp
  from 1 2 3 show ?thesis by auto
 qed  
next
  case (mTHRU_d a1 a2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a1 THRU a2)) \<equiv>\<^sub>i \<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2))" by simp
   have 2: "\<turnstile> \<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<rhd>(di(\<M> a1) \<and>\<^sub>i di(\<M> a2)))" using FstFixFst by auto
   have 3: "\<turnstile> \<rhd>(\<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2))) \<equiv>\<^sub>i \<rhd>(\<M> (a1 THRU a2))" by simp
   from 1 2 3 show ?thesis by auto
 qed
next
  case (mTHEN_d a1 a2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a1 THEN a2)) \<equiv>\<^sub>i (\<M> a1) ; (\<M> a2) " 
        by simp
   have 2: "\<turnstile> (\<M> a1) ; (\<M> a2) \<equiv>\<^sub>i \<rhd>(\<M> a1) ; \<rhd>(\<M> a2)" 
        using ChopEqvChop mTHEN_d.hyps(1) mTHEN_d.hyps(2) by blast
   have 3: "\<turnstile> \<rhd>(\<M> a1) ; \<rhd>(\<M> a2) \<equiv>\<^sub>i \<rhd>(\<rhd>(\<M> a1) ; (\<M> a2))" 
        using FstFstChopEqvFstChopFst itl_prop(30) by blast
   have 4: "\<turnstile> \<rhd>(\<rhd>(\<M> a1) ; (\<M> a2)) \<equiv>\<^sub>i \<rhd>((\<M> a1) ; (\<M> a2))" 
        using FstEqvRule LeftChopEqvChop itl_prop(30) mTHEN_d.hyps(1) by blast
   have 5: "\<turnstile> \<rhd>((\<M> a1) ; (\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<M> (a1 THEN a2))" 
        by simp
   from 1 2 3 4 5 show ?thesis by auto
  qed
next
  case (mWITH_d a x2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a WITH x2)) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i ( x2)" 
        by simp
   have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i ( x2) \<equiv>\<^sub>i \<rhd>(\<M> a) \<and>\<^sub>i ( x2)" 
        using mWITH_d.hyps by auto
   have 3: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i ( x2) \<equiv>\<^sub>i \<rhd>(\<rhd>(\<M> a) \<and>\<^sub>i ( x2))"  
        using FstFstAndEqvFstAnd itl_prop(30) by blast
   have 4: "\<turnstile> \<rhd>(\<rhd>(\<M> a) \<and>\<^sub>i ( x2)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<and>\<^sub>i ( x2))" 
        using "2" FstEqvRule itl_prop(30) by blast
   have 5: "\<turnstile> \<rhd>((\<M> a) \<and>\<^sub>i ( x2)) \<equiv>\<^sub>i \<rhd>(\<M> (a WITH x2))" 
        by simp
   from 1 2 3 4 5 show ?thesis by auto
 qed
qed

lemma MGuardFalseEqvFalse:
 "\<turnstile> \<M> (GUARD false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M>(GUARD false\<^sub>i) \<equiv>\<^sub>i \<M>(EMPTY WITH (init false\<^sub>i))" by (simp add: mGUARD_d_def)
  have 2: "\<turnstile> \<M>(EMPTY WITH (init false\<^sub>i)) \<equiv>\<^sub>i \<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i)" by simp
  have 3: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i (init false\<^sub>i)" by simp
  have 4: "\<turnstile> \<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i) \<equiv>\<^sub>i \<M>(EMPTY) \<and>\<^sub>i false\<^sub>i" using "3" by simp
  have 5: "\<turnstile> \<M>(EMPTY) \<and>\<^sub>i false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" by simp
  have 6: "\<turnstile>\<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using "4" "5" by simp
  have 7: "\<turnstile>\<M>(EMPTY WITH (init false\<^sub>i)) \<equiv>\<^sub>i false\<^sub>i" using "2" "6" by simp
  have 8: "\<turnstile>\<M>(GUARD false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using "1" "7" by simp
  from 8 show ?thesis by auto
qed

lemma MFirstFalseEqvFalse:
 "\<turnstile> \<M> (FIRST false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M>(FIRST false\<^sub>i) \<equiv>\<^sub>i \<rhd> false\<^sub>i" by simp
  have 2:  "\<turnstile> \<M>(FIRST false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using FstFalse by simp
 from 2 show ?thesis by auto
qed

lemma MFailAlt:
 "\<turnstile> \<M> FAIL \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M> FAIL \<equiv>\<^sub>i \<M> (GUARD (false\<^sub>i)) " by (simp add: mFAIL_d_def)
  have 2: "\<turnstile> \<M> (GUARD (false\<^sub>i)) \<equiv>\<^sub>i false\<^sub>i" using MGuardFalseEqvFalse by auto
 from 1 2 show ?thesis by auto
qed

lemma MFailEqvFirstFalseWithinEmpty:
 " (FAIL) \<simeq> ((FIRST false\<^sub>i) WITHIN ( empty ))"
proof -
  have 1: "\<turnstile> \<M>( (FIRST false\<^sub>i) WITHIN ( empty )) \<equiv>\<^sub>i \<M> ((FIRST false\<^sub>i) WITH (Limit empty) )" 
       by (simp add: mWITHIN_d_def)
  have 2: "\<turnstile> \<M> ((FIRST false\<^sub>i) WITH (Limit empty) ) \<equiv>\<^sub>i \<M> (FIRST false\<^sub>i) \<and>\<^sub>i (Limit empty ) " 
       by simp
  have 3: "\<turnstile> \<M> ((FIRST false\<^sub>i) WITH (Limit empty) ) \<equiv>\<^sub>i false\<^sub>i" 
       using MFirstFalseEqvFalse by auto
  have 4: "\<turnstile> \<M>( (FIRST false\<^sub>i) WITHIN ( empty )) \<equiv>\<^sub>i false\<^sub>i" 
       using "1" "3" by auto
  have 5: "\<turnstile> (\<M> FAIL) \<equiv>\<^sub>i false\<^sub>i" 
       using MFailAlt by simp
 from 4 5 show ?thesis by (metis eq_d_def itl_prop(30) prop21)
 qed

lemma MEmptyAlt:
 "\<turnstile> (\<M> EMPTY) \<equiv>\<^sub>i empty"
proof -
  have 1: "\<turnstile> (\<M> EMPTY) \<equiv>\<^sub>i (\<M> (FIRST empty))" by (simp add: mEMPTY_d_def)
  have 2: "\<turnstile> (\<M> (FIRST empty)) \<equiv>\<^sub>i \<rhd> empty" by simp
  have 3: "\<turnstile> \<rhd> empty \<equiv>\<^sub>i empty" using FstEmpty by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MSkipAlt:
 "\<turnstile> \<M> SKIP \<equiv>\<^sub>i skip"
proof -
  have 1: "\<turnstile> \<M> SKIP \<equiv>\<^sub>i \<M>(FIRST skip)" by (simp add: mSKIP_d_def)
  have 2: "\<turnstile> \<M>(FIRST skip) \<equiv>\<^sub>i \<rhd> skip" by simp
  have 3: "\<turnstile> \<rhd> skip \<equiv>\<^sub>i skip" using FstSkip by simp
 from 1 2 3 show ?thesis by auto
qed

lemma MGuardAlt:
 "\<turnstile> \<M> (GUARD(w)) \<equiv>\<^sub>i empty \<and>\<^sub>i init w"
proof -
  have 1: "\<turnstile> \<M> (GUARD(w)) \<equiv>\<^sub>i \<M>(EMPTY WITH ( (init w)))" by (simp add:mGUARD_d_def)
  have 2: "\<turnstile> \<M>(EMPTY WITH ( (init w))) \<equiv>\<^sub>i (\<M> EMPTY) \<and>\<^sub>i ( init w)" by simp
  have 3: "\<turnstile> (\<M> EMPTY) \<and>\<^sub>i ( init w) \<equiv>\<^sub>i empty \<and>\<^sub>i ( init w)" using MEmptyAlt prop06 by blast
  have 4: "\<turnstile>  empty \<and>\<^sub>i ( init w) \<equiv>\<^sub>i empty \<and>\<^sub>i init w" by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma MLengthAlt:
 "\<turnstile>\<M>(LEN(k)) \<equiv>\<^sub>i len(k)"
proof -
  have 1: "\<turnstile> \<M>(LEN(k)) \<equiv>\<^sub>i \<M>(FIRST(len(k)))" by (simp add:mLEN_d_def)
  have 2: "\<turnstile> \<M>(FIRST(len(k))) \<equiv>\<^sub>i \<rhd>(len(k))" by simp
  have 3: "\<turnstile> \<rhd>(len(k)) \<equiv>\<^sub>i len(k)" using FstLenEqvLen by blast
 from 1 2 3 show ?thesis by auto
qed

lemma MAlwaysAlt:
  "\<turnstile> \<M>(a ALWAYS w) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<box> (init w)"
proof -
  have 1: "\<turnstile> \<M>(a ALWAYS w) \<equiv>\<^sub>i \<M>(a WITH (bi (fin (init w))))" 
       by (simp add: mALWAYS_d_def)
  have 2: "\<turnstile> \<M>(a WITH (bi (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bi (fin (init w)))" 
       by simp
  have 3: "\<turnstile> \<M>(a) \<and>\<^sub>i (bi (fin (init w))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<box> (init w)" 
       using BoxStateEqvBiFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MSometimeAlt:
 "\<turnstile> \<M>(a SOMETIME w) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<diamond> (init w)"
proof -
  have 1: "\<turnstile> \<M>(a SOMETIME w) \<equiv>\<^sub>i \<M>(a WITH (di (fin (init w))))" 
       by (simp add: mSOMETIME_d_def)
  have 2: "\<turnstile>\<M>(a WITH (di (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (di (fin (init w)))" 
       by simp
  have 3: "\<turnstile>\<M>(a WITH (di (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i  \<diamond> (init w)" 
       using DiamondStateEqvDiFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MWithinAlt:
 "\<turnstile> \<M>(a WITHIN f) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bs (\<not>\<^sub>i f))"
proof -
  have 1: "\<turnstile> \<M>(a WITHIN f) \<equiv>\<^sub>i \<M>(a WITH (bs (\<not>\<^sub>i f)))" 
       by (simp add: mWITHIN_d_def mlimit_d_def)
  have 2: "\<turnstile> \<M>(a WITH (bs (\<not>\<^sub>i f))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bs (\<not>\<^sub>i f))" 
       by simp
  from 1 2 show ?thesis by simp
qed

lemma MTimesAlt:
 "\<turnstile> \<M>(a TIMES k) \<equiv>\<^sub>i power (\<M> a) k"
proof 
 (induct k)
 case 0
 then show ?case 
  proof -
   have 1: "\<turnstile> \<M>  a TIMES 0 \<equiv>\<^sub>i \<M> EMPTY" by simp
   have 2: "\<turnstile> \<M> EMPTY \<equiv>\<^sub>i empty " using MEmptyAlt by simp
   have 3: "\<turnstile> empty \<equiv>\<^sub>i power (\<M> a) 0" by simp
   from 1 2 3 show ?thesis by auto
  qed
 next
 case (Suc k)
 then show ?case 
  proof -
   have 1: "\<turnstile> \<M>  a TIMES Suc k \<equiv>\<^sub>i \<M>(a THEN (a TIMES k) ) " 
        by simp
   have 2: "\<turnstile> \<M>(a THEN (a TIMES k) ) \<equiv>\<^sub>i (\<M> a);(\<M>(a TIMES k))" 
        by simp
   have 3: "\<turnstile> (\<M> a);(\<M>(a TIMES k)) \<equiv>\<^sub>i (\<M> a);(power (\<M> a) k)" 
        using RightChopEqvChop Suc.hyps by blast
   have 4: "\<turnstile> (\<M> a);(power (\<M> a) k) \<equiv>\<^sub>i power (\<M> a) (Suc k)" 
        by simp
   from 1 2 3 4 show ?thesis by auto
  qed
 qed

lemma MUptoAlt:
 "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i ((\<M> a) \<and>\<^sub>i (\<M> b)) "
proof -
  have 1: "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))" 
       by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i (\<rhd>(\<M> a) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) " 
       using FstWithOrEqv by blast
  have 3: "\<turnstile> (\<rhd>(\<M> a) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ((\<M> b) \<or>\<^sub>i \<not>\<^sub>i(\<M> b)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i 
            ((\<M> b) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i \<not>\<^sub>i(\<M> a)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a)))"  
       using MFixFst by auto
  have 4: "\<turnstile> ((\<M> a) \<and>\<^sub>i ((\<M> b) \<or>\<^sub>i \<not>\<^sub>i(\<M> b)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i 
            ((\<M> b) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i \<not>\<^sub>i(\<M> a)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) \<equiv>\<^sub>i 
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) )" 
       by auto
  have 5: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( ((\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( (\<rhd>(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( (\<rhd>(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) )" 
       by (simp add: first_d_def)
  have 6: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( (\<rhd>(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( (\<rhd>(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) " 
       using MFixFst by auto
  have 7: "\<turnstile> (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<equiv>\<^sub>i bi(\<not>\<^sub>i(\<M> b))" 
       using AndBsEqvBi by blast
  have 8: "\<turnstile> (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<equiv>\<^sub>i bi(\<not>\<^sub>i(\<M> a))" 
       using AndBsEqvBi by blast
  have 9: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ((\<not>\<^sub>i(\<M> b)) \<and>\<^sub>i bs(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ((\<not>\<^sub>i(\<M> a)) \<and>\<^sub>i bs(\<not>\<^sub>i(\<M> a)))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> a)))) )" 
       using "7" "8" by auto
  have 10: "\<turnstile>((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> a)))) ) \<equiv>\<^sub>i 
            (( (\<M> a) \<and>\<^sub>i (\<M> b)) \<or>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> b)))) \<or>\<^sub>i 
            (( (\<M> b) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i ( (\<M> b) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> a)))) "  
       by auto
  have 11: "\<turnstile> (( (\<M> a) \<and>\<^sub>i (\<M> b)) \<or>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> b)))) \<or>\<^sub>i 
            (( (\<M> b) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i ( (\<M> b) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> a)))) \<equiv>\<^sub>i
           ( (\<M> a) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i ((\<M> a) \<and>\<^sub>i (\<M> b)) " 
       by auto
 from 1 2 3 4 5 6 9 10 11 show ?thesis by auto
qed

lemma MThruAlt:
 "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i di(\<M> a))"
proof -
  have 1: "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by simp
  have 2: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i (\<rhd>(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i di(\<M> a))" 
       using FstDiAndDiEqv by auto
  have 3: "\<turnstile> (\<rhd>(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i di(\<M> a))" 
       using MFixFst by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MHaltAlt:
 "\<turnstile> \<M>(HALT w) \<equiv>\<^sub>i halt(init w)"
proof -
  have 1: "\<turnstile> \<M>(HALT w) \<equiv>\<^sub>i \<M>(FIRST (fin (init  w)))" by (simp add: mHALT_d_def)
  have 2: "\<turnstile> \<M>(FIRST (fin (init  w))) \<equiv>\<^sub>i \<rhd> (fin (init  w))" by simp
  have 3: "\<turnstile> \<rhd> (fin (init  w)) \<equiv>\<^sub>i halt(init w)" using HaltStateEqvFstFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MFailUpto:
" ( (FAIL UPTO a)) \<simeq> (a)"
proof -
  have 1: "\<turnstile> (\<M> (FAIL UPTO a)) \<equiv>\<^sub>i \<rhd>((\<M> FAIL) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> \<M> FAIL \<or>\<^sub>i \<M> a \<equiv>\<^sub>i false\<^sub>i \<or>\<^sub>i \<M> a" using MFailAlt by auto
  have 3: "\<turnstile> \<rhd>((\<M> FAIL) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<rhd>(false\<^sub>i \<or>\<^sub>i (\<M> a))" using 2 FstEqvRule by blast
  have 4: "\<turnstile> false\<^sub>i \<or>\<^sub>i (\<M> a) \<equiv>\<^sub>i \<M> a" by simp
  have 5: "\<turnstile> \<rhd>(false\<^sub>i \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i  \<rhd>(\<M> a)" using 4 FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i \<M> a" using MFixFst by auto
  from 1 2 3 4 5 6 show ?thesis by (simp add: eq_d_def)
qed

lemma MFailThru:
 "(FAIL THRU a) \<simeq>  FAIL"
proof -
  have 1: "\<turnstile> \<M>(FAIL THRU a) \<equiv>\<^sub>i \<rhd>(di (\<M> FAIL) \<and>\<^sub>i di(\<M> a))" 
       by simp
  have 2: "\<turnstile> \<rhd>(di (\<M> FAIL) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) " 
       using MFailAlt DiEqvDi FstEqvRule prop06 by blast
  have 3: "\<turnstile> di false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" 
       by simp
  hence 4: "\<turnstile> \<rhd>(di (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>( (false\<^sub>i) \<and>\<^sub>i di(\<M> a))" 
       using FstEqvRule prop06 by blast
  have 5: "\<turnstile> \<rhd>( (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>false\<^sub>i" 
       using FstEqvRule itl_prop(19) by blast
  have 6: "\<turnstile> \<rhd>false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" using FstFalse 
       by auto
  have 7: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" 
       using MFailAlt by auto
 from 1 2 4 5 6 7 show ?thesis by (metis eq_d_def prop03)
qed

lemma MFailAnd:
" (FAIL AND a) \<simeq>  FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL AND a) \<equiv>\<^sub>i (\<M> FAIL) \<and>\<^sub>i (\<M> a)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> FAIL) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i false\<^sub>i \<and>\<^sub>i (\<M> a)" using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> \<M> (FAIL AND a) \<equiv>\<^sub>i false\<^sub>i" using 1 2 3 by auto
  have 5: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
  from 1 2 3 4 5 show ?thesis by (metis eq_d_def itl_prop(30) prop21)
qed

lemma MThenFail:
 " (a THEN FAIL) \<simeq>  FAIL"
proof -
  have 1: "\<turnstile> \<M> (a THEN FAIL) \<equiv>\<^sub>i (\<M> a);(\<M> FAIL)" by simp
  have 2: "\<turnstile> (\<M> a);(\<M> FAIL) \<equiv>\<^sub>i (\<M> a);false\<^sub>i " using MFailAlt by auto
  have 3: "\<turnstile> (\<M> a);false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by (metis eq_d_def itl_prop(30) prop21)
qed

lemma MFailThen:
 " (FAIL THEN a) \<simeq> FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL THEN a) \<equiv>\<^sub>i (\<M> FAIL);(\<M> a)" by simp
  have 2: "\<turnstile> (\<M> FAIL);(\<M> a) \<equiv>\<^sub>i false\<^sub>i;(\<M> a) " using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i;(\<M> a) \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by (metis eq_d_def itl_prop(30) prop21)
qed

lemma MFailWith:
 " (FAIL WITH f) \<simeq>  FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL WITH f) \<equiv>\<^sub>i (\<M> FAIL) \<and>\<^sub>i f" by simp
  have 2: "\<turnstile> (\<M> FAIL) \<and>\<^sub>i f \<equiv>\<^sub>i false\<^sub>i \<and>\<^sub>i f" using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i \<and>\<^sub>i f \<equiv>\<^sub>i false\<^sub>i" by simp
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by (metis eq_d_def itl_prop(30) prop21)
qed

lemma MWithFalse:
 " (a WITH ((false\<^sub>i))) \<simeq>  FAIL"
proof -
  have 1: "\<turnstile> \<M> (a WITH (false\<^sub>i)) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i false\<^sub>i)" by simp
  have 2: "\<turnstile> ((\<M> a) \<and>\<^sub>i false\<^sub>i) \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 show ?thesis by (simp add: MonEq)
qed

lemma MWithTrue:
 " (a WITH ((true\<^sub>i))) \<simeq>  a"
proof -
  have 1: "\<turnstile> \<M> (a WITH true\<^sub>i) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i true\<^sub>i)" by (simp)
  have 2: "\<turnstile> ((\<M> a) \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i \<M> a" by simp
 from 1 2 show ?thesis by (simp add: MonEq)
qed

lemma MEmptyUpto:
 " (EMPTY UPTO a) \<simeq>  EMPTY"
proof -
  have 1: "\<turnstile> \<M> (EMPTY UPTO a) \<equiv>\<^sub>i \<rhd>((\<M> EMPTY) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> (\<M> EMPTY)  \<equiv>\<^sub>i empty " using MEmptyAlt by auto
 hence 3: "\<turnstile> (\<M> EMPTY) \<or>\<^sub>i (\<M> a) \<equiv>\<^sub>i empty \<or>\<^sub>i (\<M> a)" by auto
 hence 4: "\<turnstile>  \<rhd>((\<M> EMPTY) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i  \<rhd>(empty \<or>\<^sub>i (\<M> a))" using FstEqvRule by blast
  have 5: "\<turnstile> \<rhd>(empty \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i empty" using FstEmptyOrEqvEmpty by blast
  have 6: "\<turnstile> empty \<equiv>\<^sub>i (\<M> EMPTY)" using MEmptyAlt by auto
 from 1 4 5 6 show ?thesis by (simp add: eq_d_def)
qed

lemma MEmptyThru:
 " (EMPTY THRU a) \<simeq> (a)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY THRU a) \<equiv>\<^sub>i \<rhd>(di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a))" by simp
  have 2: "\<turnstile> di(\<M> EMPTY) \<equiv>\<^sub>i di empty " using MEmptyAlt DiEqvDi by blast
 hence 3: "\<turnstile> di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di empty \<and>\<^sub>i di(\<M> a)" by auto
 hence 4: "\<turnstile> di empty \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di(\<M> a)" by auto
  have 5: "\<turnstile> di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di(\<M> a)" using "3" "4" by auto
 hence 6: "\<turnstile> \<rhd>(di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di(\<M> a))" using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstDiEqvFst by blast
  have 8: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 6 7 8 show ?thesis by (simp add: eq_d_def)
qed

lemma MThenEmpty: 
 " (a THEN EMPTY) \<simeq> (a)"
proof -
  have 1: "\<turnstile> \<M> (a THEN EMPTY) \<equiv>\<^sub>i (\<M> a);(\<M> EMPTY)" by simp
  have 2: "\<turnstile> (\<M> a);(\<M> EMPTY) \<equiv>\<^sub>i (\<M> a); empty" using MEmptyAlt by auto
  have 3: "\<turnstile> (\<M> a); empty \<equiv>\<^sub>i (\<M> a)" using ChopEmpty by auto
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MEmptyThen: 
 " (EMPTY THEN a) \<simeq> (a)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY THEN a) \<equiv>\<^sub>i (\<M> EMPTY);(\<M> a)" by simp
  have 2: "\<turnstile> (\<M> EMPTY);(\<M> a) \<equiv>\<^sub>i empty; (\<M> a)" using MEmptyAlt by auto
  have 3: "\<turnstile> empty;(\<M> a) \<equiv>\<^sub>i (\<M> a)" using ChopEmpty by auto
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MEmptyIterate: 
 " (EMPTY ITERATE b) \<simeq> (EMPTY)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY ITERATE b) \<equiv>\<^sub>i \<M> (EMPTY WITH (\<M> b)\<^sup>\<star>)" 
       by (simp add: mITERATE_d_def)
  have 2: "\<turnstile> \<M> (EMPTY WITH (\<M> b)\<^sup>\<star>) \<equiv>\<^sub>i \<M> EMPTY \<and>\<^sub>i (\<M> b)\<^sup>\<star>" 
       by simp
  have 3: "\<turnstile> \<M> EMPTY \<and>\<^sub>i (\<M> b)\<^sup>\<star>  \<equiv>\<^sub>i empty \<and>\<^sub>i (\<M> b)\<^sup>\<star> " 
       using MEmptyAlt by auto
  have 4: "\<turnstile> empty \<and>\<^sub>i (\<M> b)\<^sup>\<star>  \<equiv>\<^sub>i empty \<and>\<^sub>i (empty \<or>\<^sub>i (((\<M> b) \<and>\<^sub>i more);(\<M> b)\<^sup>\<star>))" 
       using ChopstarEqv by auto
  have 5: "\<turnstile> empty \<and>\<^sub>i (empty \<or>\<^sub>i (((\<M> b) \<and>\<^sub>i more);(\<M> b)\<^sup>\<star>)) \<equiv>\<^sub>i empty" 
       by auto
  have 6: "\<turnstile> \<M> (EMPTY ITERATE b) \<equiv>\<^sub>i \<M> EMPTY" 
       using 1 2 3 4 5 MEmptyAlt by auto
 from 6 show ?thesis by (metis eq_d_def)
qed

lemma MIterateIdemp: 
 " (a ITERATE a) \<simeq>  a"
proof -
  have 1: "\<turnstile> \<M> (a ITERATE a) \<equiv>\<^sub>i \<M> (a WITH (\<M> a)\<^sup>\<star>)" by (simp add: mITERATE_d_def)
  have 2: "\<turnstile> \<M> (a WITH (\<M> a)\<^sup>\<star>) \<equiv>\<^sub>i \<M> a \<and>\<^sub>i (\<M> a)\<^sup>\<star>"  by simp
  have 3: "\<turnstile> \<M> a \<and>\<^sub>i (\<M> a)\<^sup>\<star> \<equiv>\<^sub>i \<rhd>(\<M> a) \<and>\<^sub>i (\<rhd>(\<M> a))\<^sup>\<star> " using MFixFst by auto
  have 4: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i (\<rhd>(\<M> a))\<^sup>\<star> \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstAndFstStarEqvFst by simp
  have 5: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i \<M> a" using MFixFst by auto
  from 1 2 3 4 5 show ?thesis using prop03 by (metis eq_d_def)
qed

lemma MUptoIdemp:
 " (a UPTO a) \<simeq>  a"
proof -
  have 1: "\<turnstile> \<M> (a UPTO a) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstEqvRule itl_prop(27) by blast
  have 3: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruIdemp:
 " ( a THRU a) \<simeq> (a)"
proof -
  have 1: "\<turnstile> \<M> ( a THRU a) \<equiv>\<^sub>i \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> a))" by simp
  have 2: "\<turnstile> \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di (\<M> a))" using FstEqvRule itl_prop(20) by blast
  have 3: "\<turnstile> \<rhd>(di (\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstDiEqvFst by blast
  have 4: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 2 3 4 show ?thesis by (simp add: eq_d_def)
qed

lemma MAndIdemp:
 "(a AND a) \<simeq> (a)"
proof -
  have 1: "\<turnstile> \<M>(a AND a) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> a)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i (\<M> a)" by auto
 from 1 2 show ?thesis by (simp add: eq_d_def)
qed

lemma MWithIdemp:
 " ( (a WITH f)  WITH f) \<simeq> (a WITH f)"
proof -
  have 1: "\<turnstile> \<M> ( (a WITH f)  WITH f) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i ( f)) \<and>\<^sub>i ( f)" by simp
  have 2: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( f)) \<and>\<^sub>i ( f) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i ( f)" by auto
  have 3: "\<turnstile> (\<M> a) \<and>\<^sub>i ( f) \<equiv>\<^sub>i \<M>(a WITH f)" by simp
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MUptoCommut:
 "(a UPTO b) \<simeq> (b UPTO a)"
proof -
  have 1: "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))" by simp
  have 2: "\<turnstile> ((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> a))" by auto
 hence 3: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> a)) " using FstEqvRule by blast
  have 4: "\<turnstile> \<rhd>((\<M> b) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<M>(b UPTO a)" by simp
 from 1 3 4 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruCommut:
 "(a THRU b) \<simeq> (b THRU a)"
proof -
  have 1: "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))" by simp
  have 2: "\<turnstile> (di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> a)) " by auto
 hence 3: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> a))" using FstEqvRule by blast
  have 4: "\<turnstile> \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<M>(b THRU a)" by simp
 from 1 3 4 show ?thesis by (simp add: eq_d_def)
qed

lemma MAndCommut:
 "(a AND b) \<simeq> (b AND a)"
proof -
  have 1: "\<turnstile> \<M>(a AND b) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> b)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> b) \<equiv>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> a) " by auto
  have 3: "\<turnstile> (\<M> b) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i \<M>(b AND a)" by (simp add: mAND_d_def)
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MWithCommut:
 "((a WITH f) WITH g) \<simeq> ((a WITH g) WITH f) "
proof -
  have 1: "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g)" by simp
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g) \<equiv>\<^sub>i \<M>((a WITH g) WITH f)" by auto
 from 1 2  show ?thesis by (simp add: eq_d_def)
qed

lemma MWithAbsorp:
 "((a WITH f) WITH g) \<simeq> (a WITH (f \<and>\<^sub>i g)) "
proof -
  have 1: "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g)" by simp
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g) \<equiv>\<^sub>i  (\<M> a) \<and>\<^sub>i (f \<and>\<^sub>i g)" by auto
 from 1 2 show ?thesis by (simp add: MonEq)
qed

lemma MUptoAssoc:
 "((a UPTO b) UPTO c) \<simeq> (a UPTO (b UPTO c))"
proof -
  have 1: "\<turnstile> \<M>((a UPTO b) UPTO c) \<equiv>\<^sub>i \<rhd>(\<M>(a UPTO b) \<or>\<^sub>i (\<M> c))" 
       by simp
  have 2: "\<turnstile> \<rhd>(\<M>(a UPTO b) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c))" 
       by simp
  have 3: "\<turnstile> \<rhd>(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>(((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c))"  
       using FstFstOrEqvFstOrL by blast
  have 4: "\<turnstile> (((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i ((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c)))" 
       by auto
  hence 5: "\<turnstile> \<rhd>(((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c)))"  
       using FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> c)))" 
       using FstFstOrEqvFstOrR  itl_prop(30) by blast
  have 7: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<M>(b UPTO c))" 
       by simp
  have 8: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<M>(b UPTO c)) \<equiv>\<^sub>i \<M>(a UPTO (b UPTO c))" 
       by simp
 from 1 2 3 5 6 7 8 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruAssoc:
 "((a THRU b) THRU c) \<simeq> (a THRU (b THRU c))"
proof -
  have 1: "\<turnstile> \<M>((a THRU b) THRU c) \<equiv>\<^sub>i \<rhd>(di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<and>\<^sub>i di(\<M> c))" 
       by simp
  have 2: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di((di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       using DiEqvDiFst itl_prop(30) by blast
  have 3: "\<turnstile> di((di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b) " 
       using DiDiAndEqvDi by blast
  have 4: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b)" 
       using "2" "3" by auto
  hence 5: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b)  \<and>\<^sub>i di(\<M> c)" 
       by auto
  have 6: "\<turnstile> di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di (di(\<M> b)  \<and>\<^sub>i di(\<M> c))" 
       using DiDiAndEqvDi itl_prop(30) by blast
  have 7: "\<turnstile> di (di(\<M> b)  \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       using DiEqvDiFst by blast
  have 8: "\<turnstile> di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       using "6" "7" using prop03 by blast
  hence 9: "\<turnstile> di(\<M> a) \<and>\<^sub>i di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       by auto
  have 10: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i 
             di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c))) " 
       using "5" "9" by auto
  hence 11: "\<turnstile> \<rhd>(di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i 
             \<rhd>(di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))) "  
       using FstEqvRule by blast
  have 12: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))) \<equiv>\<^sub>i \<M>(a THRU (b THRU c)) " 
       by simp
 from 1 11 12 show ?thesis by (simp add: eq_d_def)
qed

lemma MAndAssoc:
 "((a AND b) AND c) \<simeq> (a AND (b AND c))"
proof -
  have 1: "\<turnstile> \<M>((a AND b) AND c) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> c)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> c) \<equiv>\<^sub>i \<M>(a AND (b AND c))" by (simp add: mAND_d_def)
 from 1 2 show ?thesis by (simp add: eq_d_def)
qed

lemma MThenAssoc:
 "((a THEN b) THEN c) \<simeq> (a THEN (b THEN c))"
proof -
  have 1: "\<turnstile> \<M>((a THEN b) THEN c) \<equiv>\<^sub>i ((\<M> a);(\<M> b));(\<M> c) " by simp
  have 2: "\<turnstile> ((\<M> a);(\<M> b));(\<M> c) \<equiv>\<^sub>i (\<M> a);((\<M> b);(\<M> c))" using ChopAssocB by blast
  have 3: "\<turnstile> (\<M> a);((\<M> b);(\<M> c)) \<equiv>\<^sub>i \<M>(a THEN (b THEN c))" by simp
 from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MUptoThruAbsorp:
 "(a UPTO (a THRU b)) \<simeq>  a"
proof -
  have 1: "\<turnstile> \<M>(a UPTO (a THRU b)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b) ))" 
       by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
            \<rhd>((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) ))" 
       using FstFstOrEqvFstOrR by auto
  have 3: "\<turnstile>  ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
             (((\<M> a) \<or>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       by auto
  have 4: "\<turnstile> (((\<M> a) \<or>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
            ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using OrDiEqvDi by auto
  have 5: "\<turnstile> ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
            ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using "3" "4" by auto
  hence 6: "\<turnstile> \<rhd> ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
             \<rhd> ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) " 
       using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd> ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
            (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) " 
       by (simp add: first_d_def)
  have 8: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i 
            (di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by auto
  hence 9: "\<turnstile> \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
            \<not>\<^sub>i((di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using prop01 by blast
  have 10: "\<turnstile> \<not>\<^sub>i((di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             \<not>\<^sub>i(((\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using AndDiEqv by auto
  have 11: "\<turnstile> \<not>\<^sub>i(((\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
              \<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by auto
  have 12: "\<turnstile> \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              \<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       using "9" "10" "11" by auto
  hence 13: "\<turnstile> bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              bs (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using BsEqvRule by blast
  have 14: "\<turnstile> bs ((\<not>\<^sub>i(\<M> a)) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using BsAndEqv using itl_prop(30) by blast
  have 141: "\<turnstile> bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using "13" "14" by auto
  hence 15: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
              bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       by auto
  have 16: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             (bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       by auto
  have 17: "\<turnstile> (bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             (\<rhd>(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       using FstEqvBsNotAndDi itl_prop(30) prop06 by blast
  have 18: "\<turnstile> (\<rhd>(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
             ((\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using MFixFst itl_prop(30) prop06 by blast
  have 19: "\<turnstile> ((\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       by auto
  have 20: "\<turnstile>   (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (\<not>\<^sub>i(di(\<M> a)) \<or>\<^sub>i \<not>\<^sub>i(di(\<M> b)) )"  
       by auto
  have 21: "\<turnstile> (\<not>\<^sub>i(di(\<M> a)) \<or>\<^sub>i \<not>\<^sub>i(di(\<M> b)) ) \<equiv>\<^sub>i ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b)) )" 
       by auto
  have 22: "\<turnstile> (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b)))  " 
       using "20" "21" by auto
  hence 23: "\<turnstile> bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i bs ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b))) " 
       using BsEqvRule by blast
  have 24: "\<turnstile> bs ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))" 
       using BsOrBsEqvBsBiOrBi  itl_prop(30) by blast
  have 25: "\<turnstile> bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))" 
       using "23" "24" by auto
  hence 26: "\<turnstile> (\<M> a) \<and>\<^sub>i bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              (\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b)))" 
       by auto
  have 27: "\<turnstile>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i 
            \<rhd>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) " 
       using MFixFst prop06 by blast 
  have 28: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i
             (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b)))" 
       by (simp add: first_d_def)
  have 29: "\<turnstile> (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i
             (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a)" 
       by auto
  have 30: "\<turnstile> (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<equiv>\<^sub>i \<rhd>(\<M> a)" 
       by (simp add: first_d_def)
  have 31: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" 
       using MFixFst by auto
  have 32: "\<turnstile> \<M>(a UPTO (a THRU b)) \<equiv>\<^sub>i 
            (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using "1" "2" "6" "7" by auto
  have 33: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using "15" "16" "17" "18" "19" by auto
  have 34: "\<turnstile> ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (\<M> a) " 
       using "26" "27" "28" "29" "30" "31" using prop03 by blast
 from 32 33 34 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruUptoAbsorp:
 "(a THRU (a UPTO b)) \<simeq> (a) "
proof -
  have 1: "\<turnstile> \<M>(a THRU (a UPTO b)) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))))" 
       by simp
  have 2: "\<turnstile>  \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)))) \<equiv>\<^sub>i 
              \<rhd>(di(\<M> a) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> b))))" 
       using DiEqvDiFst using FstEqvRule itl_prop(30) prop05 by blast
  have 3: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> b)))) \<equiv>\<^sub>i 
            \<rhd>(di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using DiOrEqv using FstEqvRule prop05 by blast
  have 4: "\<turnstile> (di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (di(\<M> a))" 
        by auto
  hence 5: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i \<rhd>(di(\<M> a))" 
       using FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" 
       using FstDiEqvFst by blast
  have 7: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" 
       using MFixFst by auto
 from 1 2 3 5 6 7 show ?thesis by (simp add: eq_d_def)
qed

lemma MUptoThruDistrib:
 "(a UPTO (b THRU c)) \<simeq> ((a UPTO b) THRU (a UPTO c))"
proof -
  have 1: "\<turnstile> \<M>((a UPTO b) THRU (a UPTO c)) \<equiv>\<^sub>i 
            \<rhd>( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) " 
       by simp
  have 2: "\<turnstile> ( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            ( di(((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> c))) )" 
       using DiEqvDiFst by (metis itl_prop(31) prop22)
  have 3: "\<turnstile> ( di(((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            ( di(\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> c))"  
       using DiOrEqv by auto
  have 4: "\<turnstile> ( di(\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i
           di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) " 
       by auto
  have 5: "\<turnstile> ( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c))" 
        using "2" "3" "4" by auto
  hence 6: "\<turnstile> \<rhd>( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            \<rhd>( di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) " 
       using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>( di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) \<equiv>\<^sub>i
            \<rhd>( \<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) " 
       using FstFstOrEqvFstOr by auto
  have 8: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>((\<M> a))" 
       using FstDiEqvFst  by blast
  have 9: "\<turnstile> \<rhd>((\<M> a)) \<equiv>\<^sub>i (\<M> a) " 
       using MFixFst by auto
  have 10: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i (\<M> a)" 
       using "8" "9" by auto
  hence 11: "\<turnstile> \<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i 
              (\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))" 
       by auto
  hence 12: "\<turnstile> \<rhd>(\<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i 
              \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)))" 
       using FstEqvRule by blast
  have 13: "\<turnstile>\<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i \<M>(a UPTO (b THRU c)) " 
       by simp
 from 1 6 7 12 13 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruUptoDistrib:
 "(a THRU (b UPTO c)) \<simeq> ((a THRU b) UPTO (a THRU c))"
proof -
  have 1: "\<turnstile> \<M>((a THRU b) UPTO (a THRU c)) \<equiv>\<^sub>i
          \<rhd>(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> c)))" by simp
  have 2: "\<turnstile> \<rhd>(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            \<rhd>((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c)))" using FstFstOrEqvFstOr by auto
  have 3: "\<turnstile> ((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i (di(\<M> b) \<or>\<^sub>i di(\<M> c)))" by auto
  have 4: "\<turnstile> (di(\<M> a) \<and>\<^sub>i (di(\<M> b) \<or>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di( (\<M> b) \<or>\<^sub>i (\<M> c)))" using DiOrEqv by auto
  have 5: "\<turnstile> (di(\<M> a) \<and>\<^sub>i di( (\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c))))" using DiEqvDiFst prop05 by blast
  have 6: "\<turnstile> ((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c))))" using "3" "4" "5" by auto
 hence 7: "\<turnstile> \<rhd>((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c)))) " using FstEqvRule by blast
  have 8: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c)))) \<equiv>\<^sub>i
            \<M>(a THRU (b UPTO c))" by simp
 from 1 2 7 8 show ?thesis by (simp add: eq_d_def)
qed

lemma MThruUptoRDistrib:
 "((a THRU b) UPTO c) \<simeq> ((a UPTO c) THRU (b UPTO c))"
proof -
  have 1: "((a THRU b) UPTO c) \<simeq> (c UPTO (a THRU b))" 
       using MUptoCommut by auto
  have 2: "(c UPTO (a THRU b)) \<simeq> ((c UPTO a) THRU (c UPTO b))" 
       using MUptoThruDistrib by auto
  have 3: "(c UPTO a) \<simeq> (a UPTO c)" 
       using MUptoCommut by auto
  have 4: "(c UPTO b) \<simeq> (b UPTO c)" 
       using MUptoCommut by auto
  have 5:  "((c UPTO a) THRU (c UPTO b)) \<simeq> ((a UPTO c) THRU (c UPTO b))" 
       using 3  by (simp add: MonEqRefl MonEqSubstThru) 
  have 6:  "((a UPTO c) THRU (c UPTO b)) \<simeq> ((c UPTO b) THRU (a UPTO c))" 
       using MThruCommut by auto
  have 7:  "((c UPTO b) THRU (a UPTO c)) \<simeq> ((b UPTO c) THRU (a UPTO c))" 
       using 4 by (simp add: MonEqRefl MonEqSubstThru)
  from 1 2 5 6 7 show ?thesis using MThruCommut by (metis MonEqTrans)
qed

lemma MUptoThruRDistrib:
 "((a UPTO b) THRU c) \<simeq> ((a THRU c) UPTO (b THRU c))"
proof -
  have 1: "((a UPTO b) THRU c) \<simeq> (c THRU (a UPTO b))" 
       using MThruCommut by auto
  have 2: "(c THRU (a UPTO b)) \<simeq> ((c THRU a) UPTO (c THRU b))" 
       using MThruUptoDistrib by auto
  have 3: "(c THRU a) \<simeq> (a THRU c)" 
       using MThruCommut by auto
  have 4: "(c THRU b) \<simeq> (b THRU c)" 
       using MThruCommut by auto
  have 5:  "((c THRU a) UPTO (c THRU b)) \<simeq> ((a THRU c) UPTO (c THRU b))" 
       using 3 by (simp add: MonEqRefl MonEqSubstUpto)
  have 6:  "((a THRU c) UPTO (c THRU b)) \<simeq> ((c THRU b) UPTO (a THRU c))" 
       using MUptoCommut by auto
  have 7:  "((c THRU b) UPTO (a THRU c)) \<simeq> ((b THRU c) UPTO (a THRU c))" 
       using 4 by (simp add: MonEqRefl MonEqSubstUpto)
  from 1 2 5 6 7 show ?thesis using MUptoCommut by (metis MonEqTrans)
qed


lemma MWithAndDistrib:
 "((a AND b) WITH f) \<simeq> ((a WITH f) AND (b WITH f))"
proof -
  have 1: "\<turnstile> \<M>((a AND b) WITH f) \<equiv>\<^sub>i \<M>(a AND b) \<and>\<^sub>i f" 
       by simp
  have 2: "\<turnstile> \<M>(a AND b) \<equiv>\<^sub>i \<M>(a WITH (\<M> b))" 
       by (simp add: mAND_d_def)
  have 3: "\<turnstile> \<M>(a AND b) \<and>\<^sub>i f \<equiv>\<^sub>i \<M>(a WITH (\<M> b)) \<and>\<^sub>i f" 
       using 2 prop06 by simp
  have 4: "\<turnstile> \<M>(a WITH (\<M> b)) \<and>\<^sub>i f \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<M>(b) \<and>\<^sub>i f" 
       by simp
  have 5: "\<turnstile> \<M>(a) \<and>\<^sub>i \<M>(b) \<and>\<^sub>i f \<equiv>\<^sub>i (\<M>(a) \<and>\<^sub>i f) \<and>\<^sub>i (\<M>(b) \<and>\<^sub>i f)" 
       by auto
  have 6: "\<turnstile> (\<M>(a) \<and>\<^sub>i f) \<and>\<^sub>i (\<M>(b) \<and>\<^sub>i f) \<equiv>\<^sub>i \<M>(a WITH f) \<and>\<^sub>i \<M>(b WITH f)" 
       by simp
  have 7: "\<turnstile> \<M>(a WITH f) \<and>\<^sub>i \<M>(b WITH f) \<equiv>\<^sub>i \<M>((a WITH f) WITH (\<M>(b WITH f)))" 
       by simp
  have 8: "\<turnstile> \<M>((a WITH f) WITH (\<M>(b WITH f))) \<equiv>\<^sub>i \<M>((a WITH f) AND (b WITH f))" 
       by (simp add: mAND_d_def)
  from 1 2 3 4 5 6 7 8 show ?thesis by (simp add: eq_d_def)
qed


lemma MHaltWithAndDistrib:
 "(((HALT w) WITH f) AND ((HALT w) WITH g)) \<simeq> ((HALT w) WITH (f \<and>\<^sub>i g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) AND ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<M>(((HALT w) WITH f) WITH (\<M>((HALT w) WITH g)))" 
       by (simp add: mAND_d_def)
  have 2: "\<turnstile> \<M>(((HALT w) WITH f) WITH (\<M>((HALT w) WITH g))) \<equiv>\<^sub>i 
             \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i \<M>(HALT w) \<and>\<^sub>i g" 
       by (simp add: mHALT_d_def)
  have 3: "\<turnstile> \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i \<M>(HALT w) \<and>\<^sub>i g \<equiv>\<^sub>i \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g"  
       by auto
  from 1 2 3 show ?thesis by (simp add: eq_d_def)
qed

lemma MHaltWithUptoHaltWithEqvHaltWithOr:
 "(((HALT w) WITH f) UPTO ((HALT w) WITH g)) \<simeq> ((HALT w) WITH (f \<or>\<^sub>i g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) UPTO ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>(\<M>((HALT w) WITH f) \<or>\<^sub>i \<M>((HALT w) WITH g))" 
       by simp
  have 2: "\<turnstile> \<rhd>(\<M>((HALT w) WITH f) \<or>\<^sub>i \<M>((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>((\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g))" 
       by simp
  have 3: "\<turnstile> (\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g))" 
       by auto 
  have 4: "\<turnstile> \<rhd>((\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g))" 
       using 3 FstEqvRule by blast
  have 5: "\<turnstile> \<rhd>(\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<or>\<^sub>i g)))" 
       by simp
  have 6: "\<turnstile> (\<M>((HALT w) WITH (f \<or>\<^sub>i g))) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<or>\<^sub>i g)))" 
       using MFixFst by blast
  from 1 2 3 4 5 6 show ?thesis by (simp add: eq_d_def)
qed

lemma MHaltWithThruHaltWithEqvHaltWithAndHaltWith:
 "(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<simeq> (((HALT w) WITH f) AND ((HALT w) WITH g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>( di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) )" 
       by simp
  have 2: "\<turnstile> di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(halt(init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt(init w) \<and>\<^sub>i g)" 
       using MHaltAlt DiEqvDi by auto
  have 3: "\<turnstile> di(halt(init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt(init w) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(halt(init w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using DiHaltAndDiHaltAndEqvDiHaltAndAnd by simp
  have 4: "\<turnstile> di( halt(init w) \<and>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using MHaltAlt by auto
  have 5: "\<turnstile> di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using 2 3 4 by simp
  have 6: "\<turnstile> \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g))" 
       using 5 FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using FstDiEqvFst by simp
  have 8: "\<turnstile> \<rhd>(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<and>\<^sub>i g)))" 
       by simp
  have 9: "\<turnstile> \<M>((HALT w) WITH (f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<and>\<^sub>i g)))" 
       using MFixFst by blast
  have 10: "\<turnstile> \<M>(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<and>\<^sub>i g))" 
       using 1 2 3 4 5 6 7 8 9 by simp
  have 11: "\<turnstile> \<M>(((HALT w) WITH f) AND ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<and>\<^sub>i g))" 
       using MHaltWithAndDistrib using eq_d_def by blast
  have 12: "\<turnstile> \<M>((HALT w) WITH (f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<M>(((HALT w) WITH f) AND ((HALT w) WITH g))" 
       using 11 by simp
  from 10 12 show ?thesis by (simp add: eq_d_def)
qed

lemma MThenAndDistrib:
 "(a THEN (b AND c)) \<simeq> ((a THEN b) AND (a THEN c))"
proof -
  have 1: "\<turnstile> \<M>(a THEN (b AND c)) \<equiv>\<^sub>i (\<M>(a)) ; (\<M>(b AND c))" 
       by simp
  have 2: "\<turnstile> (\<M>(a)) ; (\<M>(b AND c)) \<equiv>\<^sub>i (\<M>(a)) ; (\<M>(b) \<and>\<^sub>i \<M>(c))"
       by (simp add: mAND_d_def)
  have 3: "\<turnstile> (\<M>(a)) ; (\<M>(b) \<and>\<^sub>i \<M>(c)) \<equiv>\<^sub>i  \<rhd> (\<M>(a)) ; (\<M>(b) \<and>\<^sub>i \<M>(c))"
       using MFixFst LeftChopEqvChop by blast
  have 4: "\<turnstile> \<rhd> (\<M>(a)) ; (\<M>(b) \<and>\<^sub>i \<M>(c)) \<equiv>\<^sub>i  ((\<rhd> (\<M>(a)) ; (\<M>(b))) \<and>\<^sub>i (\<rhd> (\<M>(a)) ; (\<M>(c))))"
       using LFstAndDistrC by fastforce
  have 5: "\<turnstile>((\<rhd> (\<M>(a)) ; (\<M>(b))) \<and>\<^sub>i (\<rhd> (\<M>(a)) ; (\<M>(c)))) \<equiv>\<^sub>i 
            (( (\<M>(a)) ; (\<M>(b))) \<and>\<^sub>i ( (\<M>(a)) ; (\<M>(c))) ) " 
    using MFixFst by auto
  have 6: "\<turnstile> (( (\<M>(a)) ; (\<M>(b))) \<and>\<^sub>i ( (\<M>(a)) ; (\<M>(c))) ) \<equiv>\<^sub>i
             (\<M>(a THEN b) \<and>\<^sub>i \<M>(a THEN c))  " 
       by simp
  have 7: "\<turnstile> (\<M>(a THEN b) \<and>\<^sub>i \<M>(a THEN c)) \<equiv>\<^sub>i \<M>((a THEN b) AND (a THEN c))"
        by (simp add: mAND_d_def) 
  from 1 2 3 4 5 6 7 show ?thesis using MonEq by (simp add: eq_d_def)
qed


lemma MThenUptoDistrib:
 "(a THEN (b UPTO c)) \<simeq> ((a THEN b) UPTO (a THEN c))"
proof -
 have 1: "\<turnstile> (\<M> (a THEN (b UPTO c))) \<equiv>\<^sub>i ((\<M> a);(\<rhd>((\<M> b) \<or>\<^sub>i (\<M> c))))" 
      by simp
 have 2: "\<turnstile> ((\<M> a);(\<rhd>((\<M> b) \<or>\<^sub>i (\<M> c)))) \<equiv>\<^sub>i (\<rhd>(\<M> a);(\<rhd>((\<M> b) \<or>\<^sub>i (\<M> c))))" 
      using MFixFst LeftChopEqvChop by blast
 have 3: "\<turnstile> (\<rhd>(\<M> a);(\<rhd>((\<M> b) \<or>\<^sub>i (\<M> c)))) \<equiv>\<^sub>i ((\<rhd>(\<rhd>(\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c)))))" 
      using FstFstChopEqvFstChopFst by fastforce
 have 4: "\<turnstile> \<rhd>(\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i (\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c))" 
      using MFixFst LeftChopEqvChop itl_prop(30) by blast
 have 5: "\<turnstile> (\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i ((\<M> a);(\<M> b) \<or>\<^sub>i (\<M> a);(\<M> c))" 
      using ChopOrEqv by blast
 have 6: "\<turnstile> ((\<M> a);(\<M> b) \<or>\<^sub>i (\<M> a);(\<M> c)) \<equiv>\<^sub>i (\<M>(a THEN b) \<or>\<^sub>i \<M>(a THEN c))" 
      by simp
 have 7: "\<turnstile> \<rhd>(\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i (\<M>(a THEN b) \<or>\<^sub>i \<M>(a THEN c))  "
      using "6" "5" "4" by fastforce
 have 8: "\<turnstile> \<rhd>(\<rhd>(\<M> a);((\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i \<rhd>(\<M>(a THEN b) \<or>\<^sub>i \<M>(a THEN c)) "
      using "7" FstEqvRule by blast
 have 9: "\<turnstile> \<rhd>(\<M>(a THEN b) \<or>\<^sub>i \<M>(a THEN c)) \<equiv>\<^sub>i \<M>((a THEN b) UPTO (a THEN c)) "
      by simp
 from 9 7 1 2 3 show ?thesis by (meson "8" eq_d_def prop03)
qed

lemma MThenThruDistrib:
 "(a THEN (b THRU c)) \<simeq> ((a THEN b) THRU (a THEN c))"
proof -
 have 1: "\<turnstile> \<M>(a THEN (b THRU c)) \<equiv>\<^sub>i (\<M> a);\<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))"
      by simp
 have 2: "\<turnstile> (\<M> a);\<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i \<rhd>(\<M> a);\<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))"
      using MFixFst LeftChopEqvChop by blast
 have 3: "\<turnstile> \<rhd>(\<M> a);\<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i \<rhd>(\<rhd>(\<M> a);(di(\<M> b) \<and>\<^sub>i di(\<M> c) ))"
      using FstFstChopEqvFstChopFst by fastforce
 have 4: "\<turnstile> \<rhd>(\<M> a);(di(\<M> b) \<and>\<^sub>i di(\<M> c) ) \<equiv>\<^sub>i (\<rhd>(\<M> a);di(\<M> b) \<and>\<^sub>i \<rhd>(\<M> a);di(\<M> c))" 
      using LFstAndDistrC using itl_prop(30) by blast
 have 5: "\<turnstile> (\<rhd>(\<M> a);di(\<M> b) \<and>\<^sub>i \<rhd>(\<M> a);di(\<M> c)) \<equiv>\<^sub>i ((\<M> a);di(\<M> b) \<and>\<^sub>i (\<M> a);di(\<M> c))" 
      using MFixFst by auto
 have 6: "\<turnstile> (\<M> a);di(\<M> b) \<equiv>\<^sub>i (\<M> a);((\<M> b);true\<^sub>i)"
      by (simp add: di_d_def) 
 have 7: "\<turnstile> (\<M> a);((\<M> b);true\<^sub>i) \<equiv>\<^sub>i ((\<M> a);(\<M> b));true\<^sub>i"
      using ChopAssoc by blast
 have 8: "\<turnstile> ((\<M> a);(\<M> b));true\<^sub>i \<equiv>\<^sub>i di((\<M> a);(\<M> b))"
      by (simp add: di_d_def) 
 have 9: "\<turnstile> (\<M> a);di(\<M> b) \<equiv>\<^sub>i di((\<M> a);(\<M> b))"
      using "8" "7" "6" by fastforce
 have 10: "\<turnstile> (\<M> a);di(\<M> c) \<equiv>\<^sub>i (\<M> a);((\<M> c);true\<^sub>i)"
      by (simp add: di_d_def) 
 have 11: "\<turnstile> (\<M> a);((\<M> c);true\<^sub>i) \<equiv>\<^sub>i ((\<M> a);(\<M> c));true\<^sub>i"
      using ChopAssoc by blast
 have 12: "\<turnstile> ((\<M> a);(\<M> c));true\<^sub>i \<equiv>\<^sub>i di((\<M> a);(\<M> c))"
      by (simp add: di_d_def) 
 have 13: "\<turnstile> (\<M> a);di(\<M> c) \<equiv>\<^sub>i di((\<M> a);(\<M> c))"
      using "12" "11" "10" by fastforce
 have 14: "\<turnstile> ((\<M> a);di(\<M> b) \<and>\<^sub>i (\<M> a);di(\<M> c)) \<equiv>\<^sub>i (di((\<M> a);(\<M> b)) \<and>\<^sub>i di((\<M> a);(\<M> c))) "
      using "13" "9" by fastforce
 have 15: "\<turnstile> (di((\<M> a);(\<M> b)) \<and>\<^sub>i di((\<M> a);(\<M> c))) \<equiv>\<^sub>i (di(\<M>(a THEN b)) \<and>\<^sub>i di(\<M>(a THEN c)))"
      by simp
 have 16: "\<turnstile> \<rhd>(\<M> a);(di(\<M> b) \<and>\<^sub>i di(\<M> c) ) \<equiv>\<^sub>i (di(\<M>(a THEN b)) \<and>\<^sub>i di(\<M>(a THEN c))) "  
      using "15" "14" "4" "5" by fastforce
 have 17: "\<turnstile> \<rhd>(\<rhd>(\<M> a);(di(\<M> b) \<and>\<^sub>i di(\<M> c) )) \<equiv>\<^sub>i \<rhd>(di(\<M>(a THEN b)) \<and>\<^sub>i di(\<M>(a THEN c)))" 
      using "16" FstEqvRule by blast  
 have 18: "\<turnstile> \<rhd>(di(\<M>(a THEN b)) \<and>\<^sub>i di(\<M>(a THEN c))) \<equiv>\<^sub>i \<M>((a THEN b) THRU (a THEN c))"
      by simp
 from 18 16 1 2 3 show ?thesis by (meson "17" eq_d_def prop03)
qed


end
